<html>
    <head><title>Confirmation de présence</title></head>
    <body>
    <style>
        body{
            font-family: sans-serif;
            font-size: 0.9em;
        }
        div.title{ 
            font-size: 1.4em;
            font-weight: bold;
            padding: 20px 0px 0px 0px; }
        a{
            color: SteelBlue;
            text-decoration:none;
        }
        hr{
            border: 1px solid #eeeeee
         }
        .message{ padding: 0px 0px 0px 10px; }
        .powered{ margin: 10px 0px 0px 0px; font-size:0.8em;color:lightgray;text-align:left;}
        .footer{ margin: 0px 0px 0px 0px; font-size:0.8em;color:gray;}

    </style>

    <div class='title'>Confirmation de présence</div>
    <div class='message'>
        <p>Votre présence a bien été <b>prise en compte</b> par le gestionnaire de présence.</p>

        <p>Ce message est automatique et ne prend pas de réponse.
         Si vous n'avez pas effectué de déclaration de présence et que vous recevez ce message, merci d'ignorer celui-ci.</p>
    </div>

    <div class='title'>Confirmation of attendance.</div>
    <div class='message'>
        <p>Your attendance has been <b>validated</b> by the attendance manager.</p>

        <p>This message is automatic and takes no response.
           If you have not made a declaration of presence and you receive this message, please ignore this one.</p>
    </div>


    <hr>
    <div class='powered'>Powered by Fidle/TimeSheetManager - Version : ${fidle_version}</div>
    </body>
</html>